from typing import List, Optional

from presidio_evaluator import InputSample
from presidio_evaluator.models import BaseModel


class MockModel(BaseModel):

    def predict(self, sample: InputSample) -> List[str]:
        pass


class MockTokensModel(BaseModel):
    """
    Simulates a real model, returns the prediction given in the constructor
    """

    def __init__(
        self,
        prediction: Optional[List[str]],
        entities_to_keep: List = None,
        verbose: bool = False,
        **kwargs
    ):
        super().__init__(entities_to_keep=entities_to_keep, verbose=verbose, **kwargs)
        self.prediction = prediction

    def predict(self, sample: InputSample) -> List[str]:
        return self.prediction


class IdentityTokensMockModel(BaseModel):
    """
    Simulates a real model, always return the label as prediction
    """

    def __init__(self, verbose: bool = False):
        super().__init__(verbose=verbose)

    def predict(self, sample: InputSample) -> List[str]:
        return sample.tags


class FiftyFiftyIdentityTokensMockModel(BaseModel):
    """
    Simulates a real model, returns the label or no predictions (list of 'O')
    alternately
    """

    def __init__(self, entities_to_keep: List = None, verbose: bool = False):
        super().__init__(entities_to_keep=entities_to_keep, verbose=verbose)
        self.counter = 0

    def predict(self, sample: InputSample) -> List[str]:
        self.counter += 1
        if self.counter % 2 == 0:
            return sample.tags
        else:
            return ["O" for i in range(len(sample.tags))]



from .presidio_data_generator import PresidioDataGenerator
from .presidio_pseudonymize import PresidioPseudonymization


def read_synth_dataset():
    raise DeprecationWarning(
        "read_synth_dataset is deprecated. " "Please use InputSample.read_dataset_json"
    )


__all__ = ["PresidioDataGenerator", "PresidioPseudonymization", "read_synth_dataset"]



from abc import ABC, abstractmethod
from typing import List

from presidio_evaluator import InputSample


class DatasetFormatter(ABC):
    @abstractmethod
    def to_input_samples(self) -> List[InputSample]:
        """
        Translate a dataset structure into a list of documents, to be used by models and for evaluation
        :return:
        """
        pass



"""Helper scripts for calling different NER models."""
from .base_model import BaseModel
from .crf_model import CRFModel
from .presidio_analyzer_wrapper import PresidioAnalyzerWrapper
from .presidio_recognizer_wrapper import PresidioRecognizerWrapper
from .spacy_model import SpacyModel
from .stanza_model import StanzaModel
from .flair_model import FlairModel
from .flair_train import FlairTrainer

__all__ = [
    "BaseModel",
    "CRFModel",
    "PresidioRecognizerWrapper",
    "PresidioAnalyzerWrapper",
    "SpacyModel",
    "StanzaModel",
    "FlairModel",
    "FlairTrainer",
]



import json
from pathlib import Path
from typing import List, Optional, Union, Dict, Any, Tuple

import pandas as pd
import spacy
from spacy import Language
from spacy.tokens import Doc, DocBin
from spacy.training import iob_to_biluo
from spacy.symbols import ORTH
from tqdm import tqdm
from collections import Counter

from presidio_evaluator import span_to_tag, tokenize
from presidio_evaluator.data_generator.faker_extensions import (
    FakerSpansResult,
    FakerSpan,
)

PRIVY_ENTITIES = {
    "person": "PERSON",
    "name_male": "PERSON",
    "name_female": "PERSON",
    "first_name": "PERSON",
    "first_name_male": "PERSON",
    "first_name_female": "PERSON",
    "first_name_nonbinary": "PERSON",
    "last_name": "PERSON",
    "last_name_male": "PERSON",
    "last_name_female": "PERSON",

    "address": "LOCATION",
    "street_address": "LOCATION",
    "secondary_address": "LOCATION",
    "zipcode": "LOCATION",
    "building_number": "LOCATION",
    "street_name": "LOCATION",
    "airport_name": "LOCATION",
    "airport_iata": "LOCATION",  # *
    "airport_icao": "LOCATION",  # *

    "country": "LOCATION",
    "country_code": "LOCATION",
    "state": "LOCATION",
    "state_abbr": "LOCATION",
    "city": "LOCATION",

    "coordinate": "COORDINATE",  # **
    "longitude": "COORDINATE",  # **
    "latitude": "COORDINATE",  # **

    "nationality": "NRP",
    "nation_woman": "NRP",
    "nation_man": "NRP",
    "nation_plural": "NRP",
    "religion": "NRP",

    "date": "DATE_TIME",
    "date_time": "DATE_TIME",
    "date_of_birth": "DATE_TIME",
    "day_of_week": "DATE_TIME",
    "year": "DATE_TIME",
    "month": "DATE_TIME",

    "url": "URL",
    "domain_name": "URL",

    "credit_card_number": "CREDIT_CARD",
    "credit_card_expire": "DATE_TIME",

    "iban": "IBAN_CODE",
    "bban": "US_BANK_NUMBER",  # *
    "phone_number": "PHONE_NUMBER",
    "ssn": "US_SSN",
    "passport": "US_PASSPORT",  # *
    "driver_license": "US_DRIVER_LICENSE",
    "ip_address": "IP_ADDRESS",
    "itin": "US_ITIN",
    "email": "EMAIL_ADDRESS",

    "organization": "ORGANIZATION",
    "company": "ORGANIZATION",
    "airline": "ORGANIZATION",

    "job": "TITLE",
    "prefix": "TITLE",
    "prefix_male": "TITLE",
    "prefix_female": "TITLE",
    "gender": "TITLE",

    "imei": "IMEI",
    "password": "PASSWORD",
    "license_plate": "US_LICENSE_PLATE",
    "mac_address": "MAC_ADDRESS",
    "age": "AGE",

    "currency_code": "FINANCIAL",
    "aba": "FINANCIAL",
    "swift": "FINANCIAL",

    "string": "O",
    "boolean": "O",
    "color": "O",
    "random_number": "O",
    "sha1": "O",
}

PRIVY_PRESIDIO_MODEL = {
    "person": "PER",
    "name_male": "PER",
    "name_female": "PER",
    "first_name": "PER",
    "first_name_male": "PER",
    "first_name_female": "PER",
    "first_name_nonbinary": "PER",
    "last_name": "PER",
    "last_name_male": "PER",
    "last_name_female": "PER",

    "address": "LOC",
    "street_address": "LOC",
    "secondary_address": "LOC",
    "zipcode": "LOC",
    "building_number": "LOC",
    "street_name": "LOC",
    "airport_name": "LOC",
    "airport_iata": "LOC",  # *
    "airport_icao": "LOC",  # *

    "country": "LOC",
    "country_code": "LOC",
    "state": "LOC",
    "state_abbr": "LOC",
    "city": "LOC",

    "coordinate": "LOC",  # **
    "longitude": "LOC",  # **
    "latitude": "LOC",  # **

    "nationality": "NRP",
    "nation_woman": "NRP",
    "nation_man": "NRP",
    "nation_plural": "NRP",
    "religion": "NRP",

    "date": "DATE_TIME",
    "date_time": "DATE_TIME",
    "date_of_birth": "DATE_TIME",
    "day_of_week": "DATE_TIME",
    "year": "DATE_TIME",
    "month": "DATE_TIME",

    "url": "O",
    "domain_name": "O",

    "credit_card_number": "CREDIT_CARD",
    "credit_card_expire": "DATE_TIME",

    "iban": "O",
    "bban": "US_BANK_NUMBER",  # *
    "phone_number": "O",
    "ssn": "O",
    "passport": "US_PASSPORT",  # *
    "driver_license": "US_DRIVER_LICENSE",
    "ip_address": "O",
    "itin": "US_ITIN",
    "email": "O",

    "organization": "ORG",
    "company": "ORG",
    "airline": "ORG",

    "job": "TITLE",
    "prefix": "TITLE",
    "prefix_male": "TITLE",
    "prefix_female": "TITLE",
    "gender": "TITLE",

    "imei": "IMEI",
    "password": "PASSWORD",
    "license_plate": "US_LICENSE_PLATE",
    "mac_address": "MAC_ADDRESS",
    "age": "AGE",

    "currency_code": "FINANCIAL",
    "aba": "FINANCIAL",
    "swift": "FINANCIAL",

    "string": "O",
    "boolean": "O",
    "color": "O",
    "random_number": "O",
    "sha1": "O",
}

# # mapping to match CONLL and custom entities.
# PRIVY_PRESIDIO_MODEL = {
#     # entities presidio performs poorly on
#     "DATE_TIME": "DATE_TIME",
#     "US_DRIVER_LICENSE": "US_DRIVER_LICENSE",
#     "PERSON": "PER",
#     "US_PASSPORT": "US_PASSPORT",
#     "NRP": "NRP",
#     "LOCATION": "LOC",

#     # entities not supported by presidio
#     "COORDINATE": "LOC",
#     "TITLE": "TITLE",
#     "IMEI": "IMEI",
#     "PASSWORD": "PASSWORD",
#     "US_LICENSE_PLATE": "US_LICENSE_PLATE",
#     "MAC_ADDRESS": "MAC_ADDRESS",
#     "AGE": "AGE",
#     "FINANCIAL": "FINANCIAL",
# }

PRIVY_CONLL_TRANSLATOR = {
    "PERSON": "PERSON",
    "LOCATION": "LOC",
    "ORGANIZATION": "ORG",
    # "NRP": "MISC",
    # "GPE": "LOC",
}

PRIVY_ONTONOTES_TRANSLATOR = {
    "PERSON": "PERSON",
    "NRP": "NORP",
    # "STREET_ADDRESS": "FAC",
    # "LOCATION": "FAC",
    "GPE": "GPE",
    "DATE_TIME": "DATE",
    "ORGANIZATION": "ORG",
}

PRIVY_PRESIDIO_TRANSLATOR = {
    "PERSON": "PERSON",
    "NRP": "NRP",
    "STREET_ADDRESS": "LOCATION",
    "LOCATION": "LOCATION",
    "GPE": "LOCATION",
    "DATE_TIME": "DATE_TIME",
    "CREDIT_CARD": "CREDIT_CARD",
    "URL": "URL",
    "DOMAIN_NAME": "URL",
    "IBAN_CODE": "IBAN_CODE",
    # "US_BANK_NUMBER": "US_BANK_NUMBER", # not supported somehow
    "PHONE_NUMBER": "PHONE_NUMBER",
    "US_SSN": "US_SSN",
    "US_PASSPORT": "US_PASSPORT",
    "US_DRIVER_LICENSE": "US_DRIVER_LICENSE",
    "IP_ADDRESS": "IP_ADDRESS",
    "EMAIL_ADDRESS": "EMAIL_ADDRESS",
    "ORGANIZATION": "ORGANIZATION",
}

# old
SPACY_PRESIDIO_ENTITIES = {
    "ORG": "ORGANIZATION",
    "NORP": "NRP",
    "GPE": "LOCATION",
    "LOC": "LOCATION",
    "FAC": "LOCATION",
    "PERSON": "PERSON",
    "LOCATION": "LOCATION",
    "ORGANIZATION": "ORGANIZATION",
    "DATE": "DATE_TIME",
    "TIME": "DATE_TIME",
}

PRESIDIO_SPACY_ENTITIES = {
    "PERSON": "PERSON",
    "LOCATION": "LOC",
    "GPE": "GPE",
    "ORGANIZATION": "ORG",
    "DATE_TIME": "DATE",
    "NRP": "NORP",
}


class Span:
    """
    Holds information about the start, end, type and value
    of an entity in a text
    """

    def __init__(self, entity_type, entity_value, start_position, end_position):
        self.entity_type = entity_type
        self.entity_value = entity_value
        self.start_position = start_position
        self.end_position = end_position

    def intersect(self, other, ignore_entity_type: bool):
        """
        Checks if self intersects with a different Span
        :return: If intersecting, returns the number of
        intersecting characters.
        If not, returns 0
        """

        # if they do not overlap the intersection is 0
        if (
            self.end_position < other.start_position
            or other.end_position < self.start_position
        ):
            return 0

        # if we are accounting for entity type a diff type means intersection 0
        if not ignore_entity_type and (self.entity_type != other.entity_type):
            return 0

        # otherwise the intersection is min(end) - max(start)
        return min(self.end_position, other.end_position) - max(
            self.start_position, other.start_position
        )

    @classmethod
    def from_faker_span(cls, faker_span: FakerSpan) -> "Span":
        return cls(
            entity_type=faker_span.type,
            entity_value=faker_span.value,
            start_position=faker_span.start,
            end_position=faker_span.end,
        )

    def __repr__(self):
        return (
            f"Type: {self.entity_type}, "
            f"value: {self.entity_value}, "
            f"start: {self.start_position}, "
            f"end: {self.end_position}"
        )

    def __eq__(self, other):
        return (
            self.entity_type == other.entity_type
            and self.entity_value == other.entity_value
            and self.start_position == other.start_position
            and self.end_position == other.end_position
        )

    def __hash__(self):
        return hash(
            (
                "entity_type",
                self.entity_type,
                "entity_value",
                self.entity_value,
                "start_position",
                self.start_position,
                "end_position",
                self.end_position,
            )
        )

    @classmethod
    def from_json(cls, data):
        return cls(**data)


class InputSample(object):
    def __init__(
        self,
        full_text: str,
        spans: Optional[List[Span]] = None,
        masked: Optional[str] = None,
        tokens: Optional[Doc] = None,
        tags: Optional[List[str]] = None,
        create_tags_from_span=False,
        token_model_version="en_core_web_sm",
        scheme="IO",
        metadata=None,
        template_id=None,
    ):
        """
        Hold all the information needed for evaluation in the
        presidio-evaluator framework.

        :param full_text: The raw text of this sample
        :param masked: Masked/Templated version of the raw text
        :param spans: List of spans for entities
        :param create_tags_from_span: True if tags (tokens+tags) should be added
        :param scheme: IO, BIO or BILUO. Only applicable if span_to_tag=True
        :param tokens: spaCy Doc object
        :param tags: list of strings representing the label for each token,
        given the scheme
        :param token_model_version: The name of the model to use for tokenization if no tokens provided
        :param metadata: A dictionary of additional metadata on the sample,
        in the English (or other language) vocabulary
        :param template_id: Original template (utterance) of sample, in case it was generated  # noqa
        """
        if tags is None:
            tags = []
        if tokens is None:
            tokens = []
        self.full_text = full_text
        self.masked = masked
        self.spans = spans if spans else []
        self.metadata = metadata

        # generated samples have a template from which they were generated
        if not template_id and self.metadata:
            self.template_id = self.metadata.get("template_id")
        else:
            self.template_id = template_id

        if create_tags_from_span:
            tokens, tags = self.get_tags(scheme, token_model_version)
            self.tokens = tokens
            self.tags = tags
        else:
            self.tokens = tokens
            self.tags = tags

    @classmethod
    def from_faker_spans_result(
        cls,
        faker_spans_result: FakerSpansResult,
        scheme: str = "BILUO",
        create_tags_from_span: bool = True,
        **kwargs,
    ) -> "InputSample":
        """
        Translate the FakerSpansResult object to InputSample for backward compatibility
        :param faker_spans_result: A FakerSpansResult object
        :param create_tags_from_span: True if text should be tokenized according to spans
        :param scheme: Annotation scheme for tokens (BILUO, BIO, IO). Only relevant if create_tags_from_span=True
        :param kwargs: Additional kwargs for InputSample creation
        :return: InputSample
        """
        spans = [
            Span.from_faker_span(new_span) for new_span in faker_spans_result.spans
        ]
        return cls(
            full_text=faker_spans_result.fake,
            spans=spans,
            masked=faker_spans_result.template,
            create_tags_from_span=create_tags_from_span,
            scheme=scheme,
            template_id=faker_spans_result.template_id,
            **kwargs,
        )

    def __repr__(self):
        return (
            f"Full text: {self.full_text}\n"
            f"Spans: {self.spans}\n"
            f"Tokens: {self.tokens}\n"
            f"Tags: {self.tags}\n"
        )

    def to_dict(self):

        return {
            "full_text": self.full_text,
            "masked": self.masked,
            "spans": [span.__dict__ for span in self.spans],
            "tags": self.tags,
            "tokens": [t.text for t in self.tokens],
            "template_id": self.template_id,
            "metadata": self.metadata,
        }

    @classmethod
    def from_json(cls, data, **kwargs):
        if "spans" in data:
            data["spans"] = [Span.from_json(span) for span in data["spans"]]
        return cls(**data, create_tags_from_span=True, **kwargs)

    def get_tags(self, scheme="IOB", model_version="en_core_web_sm"):
        start_indices = [span.start_position for span in self.spans]
        end_indices = [span.end_position for span in self.spans]
        tags = [span.entity_type for span in self.spans]
        tokens = tokenize(self.full_text, model_version)

        labels = span_to_tag(
            scheme=scheme,
            text=self.full_text,
            tags=tags,
            starts=start_indices,
            ends=end_indices,
            tokens=tokens,
        )

        return tokens, labels

    def to_conll(self, translate_tags: bool) -> List[Dict[str, Any]]:
        """
        Turns a list of InputSample objects to a dictionary
        containing text, pos, tag, template_id and label.
        :param translate_tags: Whether to translate tags using the PRESIDIO_SPACY_ENTITIES dictionary
        :return: Dict
        """

        conll = []
        for i, token in enumerate(self.tokens):
            if translate_tags:
                label = self.translate_tag(
                    self.tags[i], PRIVY_ENTITIES, ignore_unknown=True
                )
            else:
                label = self.tags[i]
            conll.append(
                {
                    "text": token.text,
                    "pos": token.pos_,
                    "tag": token.tag_,
                    "template_id": self.template_id,
                    "label": label,
                },
            )

        return conll

    def get_template_id(self):
        if not self.template_id:
            return self.metadata.get("template_id")

    @staticmethod
    def create_conll_dataset(
        dataset: Union[List["InputSample"], List[FakerSpansResult]],
        translate_tags=False,
        # translator=PRIVY_ENTITIES,
        to_bio=True,
        token_model_version="en_core_web_sm"
    ) -> pd.DataFrame:

        if len(dataset) <= 1:
            raise ValueError("Dataset should contain multiple records")

        if isinstance(dataset[0], FakerSpansResult):
            dataset = [
                InputSample.from_faker_spans_result(
                    record, create_tags_from_span=True, scheme="BILUO", token_model_version=token_model_version
                )
                for record in tqdm(dataset, desc="Translating spans into tokens")
            ]

        conlls = []
        i = 0
        for sample in tqdm(dataset):
            if to_bio:
                sample.biluo_to_bio()
            conll = sample.to_conll(translate_tags=translate_tags)
            for token in conll:
                token["sentence"] = i
                conlls.append(token)
            i += 1

        return pd.DataFrame(conlls)

    def to_spacy(
        self, entities=None, translate_tags=True
    ) -> Tuple[str, Dict[str, List]]:
        """
        Translates an input sample into a format which can be consumed by spaCy during training.
        :param entities: Specific entities to focus on.
        :param translate_tags: Whether to translate the existing tags into spaCy tags (PERSON, LOC, GPE, ORG)
        :return: text and a dictionary containing a list of entities, e.g.
        "Bob is my name", {"entities": [(0, 3, "PERSON")]}
        """
        entities = [
            (span.start_position, span.end_position, span.entity_type)
            for span in self.spans
            if (entities is None) or (span.entity_type in entities)
        ]
        new_entities = []
        if translate_tags:
            for entity in entities:
                new_tag = self.translate_tag(
                    entity[2], PRESIDIO_SPACY_ENTITIES, ignore_unknown=True
                )
                new_entities.append((entity[0], entity[1], new_tag))
        else:
            new_entities = entities
        return self.full_text, {"entities": new_entities}

    @classmethod
    def from_spacy_doc(
        cls, doc: Doc, translate_tags: bool = True, scheme: str = "BILUO"
    ) -> "InputSample":
        if scheme not in ("BILUO", "BILOU", "BIO", "IOB"):
            raise ValueError('scheme should be one of "BILUO","BILOU","BIO","IOB"')

        spans = []
        for ent in doc.ents:
            entity_type = (
                cls.rename_from_spacy_tag(ent.label_) if translate_tags else ent.label_
            )
            span = Span(
                entity_type=entity_type,
                entity_value=ent.text,
                start_position=ent.start_char,
                end_position=ent.end_char,
            )
            spans.append(span)

        tags = [
            f"{token.ent_iob_}-{token.ent_type_}" if token.ent_iob_ != "O" else "O"
            for token in doc
        ]
        if scheme in ("BILUO", "BILOU"):
            tags = iob_to_biluo(tags)

        return cls(
            full_text=doc.text,
            masked=None,
            spans=spans,
            tokens=doc,
            tags=tags,
            create_tags_from_span=False,
            scheme=scheme,
        )

    @staticmethod
    def get_spacy(model_version="en_core_web_sm"):
        # add custom rules to spacy tokenizer to break down html, xml, sql into single tokens
        print("loading model {}".format(model_version))
        nlp = spacy.load(model_version)
        # split on these additional characters =' is for sql queries, < and > for html, xml)
        infixes = nlp.Defaults.infixes + [r'([><(=\'),"])']
        nlp.tokenizer.infix_finditer = spacy.util.compile_infix_regex(infixes).finditer

        for tagName in "html body i br p".split():
            nlp.tokenizer.add_special_case(f"<{tagName}>", [{ORTH: f"<{tagName}>"}])
            nlp.tokenizer.add_special_case(f"</{tagName}>", [{ORTH: f"</{tagName}>"}])

        for tagName in "br p".split():
            nlp.tokenizer.add_special_case(f"<{tagName}/>", [{ORTH: f"<{tagName}/>"}])

        for tagName in "html body i br p".split():
            nlp.tokenizer.add_special_case(f"<{tagName}>", [{ORTH: f"<{tagName}>"}])
            nlp.tokenizer.add_special_case(f"</{tagName}>", [{ORTH: f"</{tagName}>"}])
        return nlp

    @staticmethod
    def create_spacy_dataset(
        dataset: List["InputSample"],
        output_path: Optional[str] = None,
        entities: List[str] = None,
        sort_by_template_id: bool = False,
        translate_tags: bool = True,
        spacy_pipeline: Optional[Language] = None,
        alignment_mode: str = "expand",
    ) -> List[Tuple[str, Dict]]:
        """
        Creates a dataset which can be used to train spaCy models.
        If output_path is provided, it also saves the dataset in a spacy format.
        See https://spacy.io/usage/training#training-data

        :param dataset: List[InputSample] to create the dataset from
        :param output_path: Location for the created spacy dataset
        :param entities: List of entities to use
        :param sort_by_template_id: Whether to sort by template id (assuming the data is generated using templates)
        :param translate_tags: Whether to translate tags to spacy tags (PERSON, LOC, GPE, ORG)
        :param spacy_pipeline: The spaCy pipeline to use when creating the spaCy dataset. Default is en_core_web_sm
        :param alignment_mode: See `Doc.char_span`
        :return: a list of input samples translated to the spacy annotation structure
        [("Bob is my name, {"entities": [(0, 3, "PERSON")]})]
        """

        def template_sort(x):
            return x.metadata["template_id"]

        if sort_by_template_id:
            dataset.sort(key=template_sort)

        if not spacy_pipeline:
            spacy_pipeline = InputSample.get_spacy()

        spacy_dataset = [
            sample.to_spacy(entities=entities, translate_tags=translate_tags)
            for sample in dataset
        ]

        # Remove 'O' spans (if certain entities were ignored)
        for sample in spacy_dataset:
            if sample[1]["entities"]:
                new_entities = [
                    span for span in sample[1]["entities"] if span[2] != "O"
                ]
                sample[1]["entities"] = new_entities

        if output_path:
            db = DocBin()
            for text, annotations in spacy_dataset:
                doc = spacy_pipeline(text)
                ents = []
                for start, end, label in annotations["entities"]:
                    if start >= end:
                        print(
                            f"Span has zero or negative size, skipping. {(start, end, label)} in text={text}"
                        )
                        continue
                    if label == "O" or not label:
                        print("Skipping missing or non-entity ('O') spans")
                        continue
                    span = doc.char_span(
                        start, end, label=label, alignment_mode=alignment_mode
                    )
                    if not span:
                        print(
                            f"Skipping illegal span {(start, end, label)}, text={text[start:end]}, full text={text}"
                        )
                        continue
                    ents.append(span)
                doc.ents = ents
                db.add(doc)
            db.to_disk(output_path)

        return spacy_dataset

    @staticmethod
    def to_json(dataset: List["InputSample"], output_file: Union[str, Path]):
        """
        Save the InputSample dataset to json.
        :param dataset: list of InputSample objects
        :param output_file: path to file
        """

        examples_json = [example.to_dict() for example in dataset]

        with open("{}".format(output_file), "w+", encoding="utf-8") as f:
            json.dump(examples_json, f, ensure_ascii=False, indent=4)

    def to_spacy_doc(self):
        doc = self.tokens
        spacy_spans = []
        for span in self.spans:
            start_token = [
                token.i for token in self.tokens if token.idx == span.start_position
            ][0]
            end_token = [
                token.i
                for token in self.tokens
                if token.idx + len(token.text) == span.end_position
            ][0] + 1
            spacy_span = spacy.tokens.span.Span(
                doc, start=start_token, end=end_token, label=span.entity_type
            )
            spacy_spans.append(spacy_span)
        doc.ents = spacy_spans
        return doc

    @staticmethod
    def translate_tag(tag: str, dictionary: Dict[str, str], ignore_unknown: bool):
        has_prefix = len(tag) > 2 and tag[1] == "-"
        no_prefix = tag[2:] if has_prefix else tag
        if no_prefix in dictionary.keys():
            return (
                tag[:2] + dictionary[no_prefix] if has_prefix else dictionary[no_prefix]
            )
        else:
            if ignore_unknown:
                return "O"
            else:
                return tag

    def biluo_to_bio(self):
        new_tags = []
        for tag in self.tags:
            new_tag = tag
            has_prefix = len(tag) > 2 and tag[1] == "-"
            if has_prefix:
                if tag[0] == "U":
                    new_tag = "B" + tag[1:]
                elif tag[0] == "L":
                    new_tag = "I" + tag[1:]
            new_tags.append(new_tag)

        self.tags = new_tags

    @staticmethod
    def rename_from_spacy_tag(spacy_tag, ignore_unknown=False):
        return InputSample.translate_tag(
            spacy_tag, SPACY_PRESIDIO_ENTITIES, ignore_unknown=ignore_unknown
        )

    @staticmethod
    def rename_to_spacy_tags(tag, ignore_unknown=True):
        return InputSample.translate_tag(
            tag, PRESIDIO_SPACY_ENTITIES, ignore_unknown=ignore_unknown
        )

    def to_flair(self):
        for i, token in enumerate(self.tokens):
            return f"{token} {token.pos_} {self.tags[i]}"

    def translate_input_sample_tags(self, dictionary=None, ignore_unknown=True):
        if dictionary is None:
            dictionary = PRESIDIO_SPACY_ENTITIES

        # Translate tags
        self.tags = [
            InputSample.translate_tag(tag, dictionary, ignore_unknown=ignore_unknown)
            for tag in self.tags
        ]

        # Translate spans
        for span in self.spans:
            if span.entity_type in dictionary:
                span.entity_type = dictionary[span.entity_type]
            elif ignore_unknown:
                span.entity_value = "O"

        # Remove spans if they were changed to "O"
        self.spans = [span for span in self.spans if span.entity_type != "O"]

    @staticmethod
    def create_flair_dataset(dataset: List["InputSample"]) -> List[str]:
        flair_samples = []
        for sample in dataset:
            flair_samples.append(sample.to_flair())

        return flair_samples

    @staticmethod
    def read_dataset_json(
        filepath: Union[Path, str] = None, length: Optional[int] = None, **kwargs
    ) -> List["InputSample"]:
        """
        Reads an existing dataset, stored in json into a list of InputSample objects
        :param filepath: Path to json file
        :param length: Number of records to return (would return 0-length)
        :return: List[InputSample]
        """
        with open(filepath, "r", encoding="utf-8") as f:
            dataset = json.load(f)

        if length:
            dataset = dataset[:length]

        input_samples = [
            InputSample.from_json(row, **kwargs) for row in tqdm(dataset, desc="tokenizing input")
        ]

        return input_samples

    @classmethod
    def convert_faker_spans(cls, fake_records: List[FakerSpansResult]) -> List["InputSample"]:
        """tokenize and transform fake samples to list of InputSample objects (data structure in presidio)"""
        input_samples = [
            # InputSample.from_faker_spans_result(faker_spans_result=fake_record)
            InputSample.from_faker_spans_result(
                fake_record, create_tags_from_span=True, scheme="BILUO", token_model_version="en_core_web_sm"
            )
            for fake_record in tqdm(fake_records)
        ]
        return input_samples

    @classmethod
    def count_entities(cls, input_samples: List["InputSample"]) -> Counter:
        count_per_entity_new = Counter()
        for record in input_samples:
            for span in record.spans:
                count_per_entity_new[span.entity_type] += 1
        return count_per_entity_new.most_common()

    @classmethod
    def remove_unsupported_entities(cls, dataset: List["InputSample"], entity_mapping: Dict[str, str]) -> None:
        """Remove records with unsupported entities using passed in entity mapping translator."""
        filtered_records = []
        excluded_entities = set()

        for sample in dataset:
            supported = True
            for span in sample.spans:
                if not span.entity_type in entity_mapping.keys():
                    supported = False
                    if span.entity_type not in excluded_entities:
                        print(f"Filtering out unsupported entity {span.entity_type}")
                    excluded_entities.add(span.entity_type)
            if supported:
                filtered_records.append(sample)
        return filtered_records



import dataclasses
import json
import random
import re
from pathlib import Path
from typing import List, Optional, Union, Generator

import numpy as np
import pandas as pd
from faker import Faker
from faker.providers import BaseProvider
from pandas import DataFrame
from tqdm import tqdm


from presidio_evaluator.data_generator.faker_extensions.span_generator import SpanFaker
from presidio_evaluator.data_generator.faker_extensions import (
    FakerSpansResult,
    NationalityProvider,
    OrganizationProvider,
    UsDriverLicenseProvider,
    IpAddressProvider,
    AddressProviderNew,
    SpanGenerator,
    RecordsFaker,
    PhoneNumberProviderNew,
    AgeProvider,
)


class PresidioDataGenerator:
    def __init__(
        self,
        custom_faker: Faker = None,
        locale: Optional[List[str]] = None,
        lower_case_ratio: float = 0.05,
    ):
        """
        Fake data generator.
        Leverages Faker to create fake PII entities into predefined templates of structure: a b c {{PII}} d e f,
        e.g. "My name is {{first_name}}."
        :param custom_faker: A Faker object provided by the user
        :param locale: A locale object to create our own Faker instance if a custom one was not provided.
        :param lower_case_ratio: Percentage of names that should start with lower case

        :example:

        >>>from presidio_evaluator.data_generator import PresidioDataGenerator

        >>>sentence_templates = [
        >>>    "My name is {{name}}",
        >>>    "Please send it to {{address}}",
        >>>    "I just moved to {{city}} from {{country}}"
        >>>]


        >>>data_generator = PresidioDataGenerator()
        >>>fake_records = data_generator.generate_fake_data(
        >>>    templates=sentence_templates, n_samples=10
        >>>)

        >>>fake_records = list(fake_records)

        >>># Print the spans of the first sample
        >>>print(fake_records[0].fake)
        I just moved to North Kim from Ukraine

        >>>print(fake_records[0].spans)
        [{"value": "Ukraine", "start": 31, "end": 38, "type": "country"}, {"value": "North Kim", "start": 16, "end": 25, "type": "city"}]

        """
        if custom_faker and locale:
            raise ValueError(
                "If a custom faker is passed, it's expected to have its locales loaded"
            )

        if custom_faker:
            self.faker = custom_faker
        else:
            generator = (
                SpanGenerator()
            )  # To allow PresidioDataGenerator to return spans and not just strings
            self.faker = Faker(local=locale, generator=generator)
        self.lower_case_ratio = lower_case_ratio

    def parse(
        self, template: str, template_id: Optional[int] = None, add_spans: bool = True
    ) -> Union[FakerSpansResult, str]:
        """
        This function replaces known PII {{tokens}} in a template sentence
        with a fake value for each token and returns a sentence with fake PII.

        Examples:
            1. "My name is {{first_name_female}} {{last_name}}".
            2. "I want to increase limit on my card # {{credit_card_number}}
                for certain duration of time. is it possible?"


        :param template: str with token(s) that needs to be replaced by fake PII
        :param template_id: The identifier of the specific template
        :param add_spans: Whether to return the spans or just the fake text

        :returns: Fake sentence.

        """
        try:
            if isinstance(self.faker.factories[0], SpanGenerator):
                fake_pattern = self.faker.parse(
                    template, add_spans=add_spans, template_id=template_id
                )
            elif isinstance(self.faker, SpanFaker):
                fake_pattern = self.faker.parse(template, template_id=template_id)
            else:
                fake_pattern = self.faker.parse(template)
            if random.random() < self.lower_case_ratio:
                fake_pattern = self._lower_pattern(fake_pattern)
            return fake_pattern
        except Exception as err:
            raise AttributeError(
                f'Failed to generate fake data based on template "{template}".'
                f"You might need to add a new Faker provider! "
                f"{err}"
            )

    @staticmethod
    def read_template_file(templates_file):
        with open(templates_file) as f:
            lines = f.readlines()
            lines = [line.replace("\\n", "\n") for line in lines]
            return lines

    @staticmethod
    def _prep_templates(raw_templates):
        print("Preparing sample sentences for ingestion")

        def make_lower_case(match_obj):
            if match_obj.group() is not None:
                return match_obj.group().lower()

        templates = [
            (
                re.sub(r"\[.*?\]", make_lower_case, template.strip())
                .replace("[", "{" + "{")
                .replace("]", "}" + "}")
            )
            for template in raw_templates
        ]

        return templates

    def generate_fake_data(
        self, templates: List[str], n_samples: int
    ) -> Union[Generator[FakerSpansResult, None, None], Generator[str, None, None]]:
        """
        Generates fake PII data whenever it encounters known faker entities in a template.
        :param templates: A list of strings containing templates
        :param n_samples: Number of samples to generate
        """

        if templates:
            templates = self._prep_templates(templates)
        else:
            templates = None

        for _ in tqdm(range(n_samples), desc="Sampling"):
            template_id = random.choice(range(len(templates)))
            template = templates[template_id]
            yield self.parse(template, template_id)

    @staticmethod
    def _lower_pattern(pattern: Union[str, FakerSpansResult]):
        if isinstance(pattern, str):
            return pattern.lower()
        elif isinstance(pattern, FakerSpansResult):
            pattern.fake = pattern.fake.lower()
            for span in pattern.spans:
                span.value = str(span.value).lower()
            return pattern

    @staticmethod
    def seed(seed_value=42):
        Faker.seed(seed_value)
        random.seed(seed_value)
        np.random.seed(seed_value)

    def add_provider_alias(self, provider_name: str, new_name: str) -> None:
        """
        Adds a copy of a provider, with a different name
        :param provider_name: Name of original provider
        :param new_name: New name
        :example:
        >>>add_provider_alias(provider_name="name", new_name="person")
        >>>self.faker.person()
        """
        original = getattr(self.faker, provider_name)

        new_provider = BaseProvider(self.faker)
        setattr(new_provider, new_name, original)
        self.faker.add_provider(new_provider)

    @staticmethod
    def update_fake_name_generator_df(fake_data: pd.DataFrame) -> DataFrame:
        """
        Adapts the csv from FakeNameGenerator to fit the data generation process used here.
        :param fake_data: a pd.DataFrame with loaded data from FakeNameGenerator.com
        :return: None
        """

        def camel_to_snake(name):
            # Borrowed from https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
            name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
            return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

        def full_name(row):
            if random.random() > 0.2:
                return str(row.first_name) + " " + str(row.last_name)
            else:
                space_after_initials = " " if random.random() > 0.5 else ". "
                return (
                    str(row.first_name)
                    + " "
                    + str(row.middle_initial)
                    + space_after_initials
                    + str(row.last_name)
                )

        def name_gendered(row):
            first_name_female, prefix_female, last_name_female = (
                (row.first_name, row.prefix, row.last_name)
                if row.gender == "female"
                else ("", "", "")
            )
            first_name_male, prefix_male, last_name_male = (
                (row.first_name, row.prefix, row.last_name)
                if row.gender == "male"
                else ("", "", "")
            )
            return (
                first_name_female,
                first_name_male,
                prefix_female,
                prefix_male,
                last_name_female,
                last_name_male,
            )

        fake_data.columns = [camel_to_snake(col) for col in fake_data.columns]

        # Update some column names to fit Faker
        fake_data.rename(
            columns={"country": "country_code", "state": "state_abbr"}, inplace=True
        )

        fake_data.rename(
            columns={
                "country_full": "country",
                "name_set": "nationality",
                "street_address": "street_name",
                "state_full": "state",
                "given_name": "first_name",
                "surname": "last_name",
                "title": "prefix",
                "email_address": "email",
                "telephone_number": "phone_number",
                "telephone_country_code": "country_calling_code",
                "birthday": "date_of_birth",
                "cc_number": "credit_card_number",
                "cc_type": "credit_card_provider",
                "cc_expires": "credit_card_expire",
                "occupation": "job",
                "domain": "domain_name",
                "username": "user_name",
                "zip_code": "zipcode",
            },
            inplace=True,
        )
        fake_data["person"] = fake_data.apply(full_name, axis=1)
        fake_data["name"] = fake_data["person"]
        genderized = fake_data.apply(
            lambda x: pd.Series(
                name_gendered(x),
                index=[
                    "first_name_female",
                    "first_name_male",
                    "prefix_female",
                    "prefix_male",
                    "last_name_female",
                    "last_name_male",
                ],
            ),
            axis=1,
            result_type="expand",
        )

        # Remove credit card data, rely on Faker's as it is more realistic
        del fake_data["credit_card_number"]

        fake_data = pd.concat([fake_data, genderized], axis="columns")
        return fake_data


if __name__ == "__main__":
    PresidioDataGenerator.seed(42)

    template_file_path = Path(Path(__file__).parent, "raw_data", "templates.txt")

    # Read FakeNameGenerator data
    fake_data_df = pd.read_csv(
        Path(Path(__file__).parent, "raw_data", "FakeNameGenerator.com_3000.csv")
    )
    # Convert column names to lowercase to match patterns
    fake_data_df = PresidioDataGenerator.update_fake_name_generator_df(fake_data_df)

    # Create a RecordsFaker (Faker object which prefers samples multiple objects from one record)
    faker = RecordsFaker(records=fake_data_df, local="en_US")
    faker.add_provider(IpAddressProvider)
    faker.add_provider(NationalityProvider)
    faker.add_provider(OrganizationProvider)
    faker.add_provider(UsDriverLicenseProvider)
    faker.add_provider(AgeProvider)
    faker.add_provider(AddressProviderNew)  # More address formats than Faker
    faker.add_provider(PhoneNumberProviderNew)  # More phone number formats than Faker

    # Create Presidio Data Generator
    data_generator = PresidioDataGenerator(custom_faker=faker, lower_case_ratio=0.05)
    data_generator.add_provider_alias(provider_name="name", new_name="person")
    data_generator.add_provider_alias(
        provider_name="credit_card_number", new_name="credit_card"
    )
    data_generator.add_provider_alias(
        provider_name="date_of_birth", new_name="birthday"
    )

    sentence_templates = PresidioDataGenerator.read_template_file(template_file_path)
    fake_patterns = data_generator.generate_fake_data(
        templates=sentence_templates, n_samples=10000
    )

    # save to json
    output_file = Path(
        Path(__file__).parent.parent.parent, "data", "presidio_data_generator_data.json"
    )

    to_json = [dataclasses.asdict(pattern) for pattern in fake_patterns]
    with open("{}".format(output_file), "w+", encoding="utf-8") as f:
        json.dump(to_json, f, ensure_ascii=False, indent=2)



from collections import Counter
from typing import List, Optional, Dict
import plotly.express as px
import plotly
from wordcloud import WordCloud
from PIL import Image
import string
import numpy as np
import pandas as pd
from tqdm import tqdm
from copy import deepcopy
from pathlib import Path

from presidio_evaluator import InputSample
from presidio_evaluator.evaluation import EvaluationResult, ModelError
from presidio_evaluator.models import BaseModel


class Evaluator:
    def __init__(
        self,
        model: BaseModel,
        verbose: bool = False,
        compare_by_io=True,
        entities_to_keep: Optional[List[str]] = None,
    ):
        """
        Evaluate a PII detection model or a Presidio analyzer / recognizer

        :param model: Instance of a fitted model (of base type BaseModel)
        :param compare_by_io: True if comparison should be done on the entity
        level and not the sub-entity level
        :param entities_to_keep: List of entity names to focus the evaluator on (and ignore the rest).
        Default is None = all entities. If the provided model has a list of entities to keep,
        this list would be used for evaluation.
        """
        self.model = model
        self.verbose = verbose
        self.compare_by_io = compare_by_io
        self.entities_to_keep = entities_to_keep
        if self.entities_to_keep is None and self.model.entities:
            self.entities_to_keep = self.model.entities

    def compare(self, input_sample: InputSample, prediction: List[str]):
        """
        Compares ground truth tags (annotation) and predicted (prediction)
        :param input_sample: input sample containing list of tags with scheme
        :param prediction: predicted value for each token
        self.labeling_scheme

        """
        annotation = input_sample.tags
        tokens = input_sample.tokens

        if len(annotation) != len(prediction):
            print(
                "Annotation and prediction do not have the"
                "same length. Sample={}".format(input_sample)
            )
            return Counter(), []

        results = Counter()
        mistakes = []

        new_annotation = annotation.copy()

        if self.compare_by_io:
            new_annotation = self._to_io(new_annotation)
            prediction = self._to_io(prediction)

        # Ignore annotations that aren't in the list of
        # requested entities.
        if self.entities_to_keep:
            prediction = self._adjust_per_entities(prediction)
            new_annotation = self._adjust_per_entities(new_annotation)
        for i in range(0, len(new_annotation)):
            # skip punctuation labeled as entity
            if tokens[i].text in string.punctuation or tokens[i].text in ["\\n", "\\t", " ", "", " \\n", "\\n "]:
                continue
            results[(new_annotation[i], prediction[i])] += 1

            if self.verbose:
                print("Annotation:", new_annotation[i])
                print("Prediction:", prediction[i])
                print(results)

            # check if there was an error
            is_error = new_annotation[i] != prediction[i]
            if is_error:
                if prediction[i] == "O":
                    mistakes.append(
                        ModelError(
                            error_type="FN",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )
                elif new_annotation[i] == "O":
                    mistakes.append(
                        ModelError(
                            error_type="FP",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )
                else:
                    mistakes.append(
                        ModelError(
                            error_type="Wrong entity",
                            annotation=new_annotation[i],
                            prediction=prediction[i],
                            token=tokens[i],
                            full_text=input_sample.full_text,
                            metadata=input_sample.metadata,
                        )
                    )

        return results, mistakes

    def _adjust_per_entities(self, tags):
        if self.entities_to_keep:
            return [tag if tag in self.entities_to_keep else "O" for tag in tags]
        else:
            return tags

    @staticmethod
    def _to_io(tags):
        """
        Translates BILUO/BIO/IOB to IO - only In or Out of entity.
        ['B-PERSON','I-PERSON','L-PERSON'] is translated into
        ['PERSON','PERSON','PERSON']
        :param tags: the input tags in BILUO/IOB/BIO format
        :return: a new list of IO tags
        """
        return [tag[2:] if "-" in tag else tag for tag in tags]

    def evaluate_sample(
        self, sample: InputSample, prediction: List[str]
    ) -> EvaluationResult:
        if self.verbose:
            print("Input sentence: {}".format(sample.full_text))

        results, mistakes = self.compare(input_sample=sample, prediction=prediction)
        return EvaluationResult(results, mistakes, sample.full_text)

    def evaluate_all(self, dataset: List[InputSample]) -> List[EvaluationResult]:
        evaluation_results = []
        if self.model.entity_mapping:
            print(
                f"Mapping entity values using this dictionary: {self.model.entity_mapping}"
            )
        for sample in tqdm(dataset, desc=f"Evaluating {self.model.__class__}"):

            # Align tag values to the ones expected by the model
            self.model.align_entity_types(sample)

            # Predict
            prediction = self.model.predict(sample)

            # Remove entities not requested
            prediction = self.model.filter_tags_in_supported_entities(prediction)

            # Switch to requested labeling scheme (IO/BIO/BILUO)
            prediction = self.model.to_scheme(prediction)

            evaluation_result = self.evaluate_sample(
                sample=sample, prediction=prediction
            )
            evaluation_results.append(evaluation_result)
        print(f"Entity counts after mapping: {InputSample.count_entities(dataset)}")

        return evaluation_results

    @staticmethod
    def align_entity_types(
        input_samples: List[InputSample],
        entities_mapping: Dict[str, str] = None,
        allow_missing_mappings: bool = False,
    ) -> List[InputSample]:
        """
        Change input samples to conform with Presidio's entities
        :return: new list of InputSample
        """

        new_input_samples = input_samples.copy()

        # A list that will contain updated input samples,
        new_list = []

        for input_sample in new_input_samples:
            contains_field_in_mapping = False
            new_spans = []
            # Update spans to match the entity types in the values of entities_mapping
            for span in input_sample.spans:
                if span.entity_type in entities_mapping.keys():
                    new_name = entities_mapping.get(span.entity_type)
                    span.entity_type = new_name
                    contains_field_in_mapping = True

                    new_spans.append(span)
                else:
                    if not allow_missing_mappings:
                        raise ValueError(
                            f"Key {span.entity_type} cannot be found in the provided entities_mapping"
                        )
            input_sample.spans = new_spans

            # Update tags in case this sample has relevant entities for evaluation
            if contains_field_in_mapping:
                for i, tag in enumerate(input_sample.tags):
                    has_prefix = "-" in tag
                    if has_prefix:
                        prefix = tag[:2]
                        clean = tag[2:]
                    else:
                        prefix = ""
                        clean = tag

                    if clean in entities_mapping.keys():
                        new_name = entities_mapping.get(clean)
                        input_sample.tags[i] = "{}{}".format(prefix, new_name)
                    else:
                        input_sample.tags[i] = "O"

            new_list.append(input_sample)

        return new_list
        # Iterate on all samples

    def calculate_score(
        self,
        evaluation_results: List[EvaluationResult],
        entities: Optional[List[str]] = None,
        beta: float = 2.5,
    ) -> EvaluationResult:
        """
        Returns the pii_precision, pii_recall, f_measure either and number of records for each entity
        or for all entities (ignore_entity_type = True)
        :param evaluation_results: List of EvaluationResult
        :param entities: List of entities to calculate score to. Default is None: all entities
        :param beta: F measure beta value
        between different entity types, or to treat these as misclassifications
        :return: EvaluationResult with precision, recall and f measures
        """

        # aggregate results
        all_results = sum([er.results for er in evaluation_results], Counter())

        # compute pii_recall per entity
        entity_recall = {}
        entity_precision = {}
        n = {}
        if not entities:
            entities = list(set([x[0] for x in all_results.keys() if x[0] != "O"]))

        for entity in entities:
            # all annotation of given type
            annotated = sum([all_results[x] for x in all_results if x[0] == entity])
            predicted = sum([all_results[x] for x in all_results if x[1] == entity])
            n[entity] = annotated
            tp = all_results[(entity, entity)]

            if annotated > 0:
                entity_recall[entity] = tp / annotated
            else:
                entity_recall[entity] = np.NaN

            if predicted > 0:
                per_entity_tp = all_results[(entity, entity)]
                entity_precision[entity] = per_entity_tp / predicted
            else:
                entity_precision[entity] = np.NaN

        # compute pii_precision and pii_recall
        annotated_all = sum([all_results[x] for x in all_results if x[0] != "O"])
        predicted_all = sum([all_results[x] for x in all_results if x[1] != "O"])
        if annotated_all > 0:
            pii_recall = (
                sum(
                    [
                        all_results[x]
                        for x in all_results
                        if (x[0] != "O" and x[1] != "O")
                    ]
                )
                / annotated_all
            )
        else:
            pii_recall = np.NaN
        if predicted_all > 0:
            pii_precision = (
                sum(
                    [
                        all_results[x]
                        for x in all_results
                        if (x[0] != "O" and x[1] != "O")
                    ]
                )
                / predicted_all
            )
        else:
            pii_precision = np.NaN
        # compute pii_f_beta-score
        pii_f_beta = self.f_beta(pii_precision, pii_recall, beta)

        # aggregate errors
        errors = []
        for res in evaluation_results:
            if res.model_errors:
                errors.extend(res.model_errors)

        evaluation_result = EvaluationResult(
            results=all_results,
            model_errors=errors,
            pii_precision=pii_precision,
            pii_recall=pii_recall,
            entity_recall_dict=entity_recall,
            entity_precision_dict=entity_precision,
            n_dict=n,
            pii_f=pii_f_beta,
            n=sum(n.values()),
        )

        return evaluation_result

    @staticmethod
    def precision(tp: int, fp: int) -> float:
        return tp / (tp + fp + 1e-100)

    @staticmethod
    def recall(tp: int, fn: int) -> float:
        return tp / (tp + fn + 1e-100)

    @staticmethod
    def f_beta(precision: float, recall: float, beta: float) -> float:
        """
        Returns the F score for precision, recall and a beta parameter
        :param precision: a float with the precision value
        :param recall: a float with the recall value
        :param beta: a float with the beta parameter of the F measure,
        which gives more or less weight to precision
        vs. recall
        :return: a float value of the f(beta) measure.
        """
        if np.isnan(precision) or np.isnan(recall) or (precision == 0 and recall == 0):
            return np.nan

        return ((1 + beta ** 2) * precision * recall) / (
            ((beta ** 2) * precision) + recall
        )

    class ErrorAnalyzer:
        def __init__(self, model, results, errors, output_folder, model_name, path_to_wordcloud_image):
            self.model = model
            self.results = results
            self.errors = errors
            self.output_folder = output_folder
            self.model_name = model_name.replace("/", "-")
            self.path_to_image = path_to_wordcloud_image

        def plot_recall_precision_f2(self) -> None:
            """Plot per-entity recall and precision"""
            d = {}
            d['entity'] = deepcopy(list(self.results.entity_recall_dict.keys()))
            d['recall'] = deepcopy(list(self.results.entity_recall_dict.values()))
            d['precision'] = deepcopy(
                list(self.results.entity_precision_dict.values()))
            d['count'] = deepcopy(list(self.results.n_dict.values()))
            d['f2_score'] = [Evaluator.f_beta(precision=precision, recall=recall, beta=2.5)
                             for recall, precision in zip(d['recall'], d['precision'])]
            df = pd.DataFrame(d)
            df['model'] = self.model_name
            # self._plot(df, plot_type="Recall")
            # self._plot(df, plot_type="Precision")
            self._plot(df, plot_type="F2_Score")

            scores_output = self.output_folder / \
                f"scores-dict-{self.model_name}.json"
            df.to_csv(scores_output, index=False)

        def _plot(self, df, plot_type: str) -> None:
            fig = px.bar(df, text_auto=".2", y='entity', orientation="h",
                         x=f'{plot_type.lower()}', color='count', barmode='group', title=f"Per-entity {plot_type} for {self.model_name}")
            fig.update_layout(barmode='group', yaxis={
                'categoryorder': 'total ascending'})
            fig.update_layout(yaxis_title=f"{plot_type}", xaxis_title="PII Entity")
            fig.update_traces(textfont_size=12, textangle=0,
                              textposition="outside", cliponaxis=False)
            fig.update_layout(
                plot_bgcolor="#FFF",
                xaxis=dict(
                    title="PII entity",
                    linecolor="#BCCCDC",  # Sets color of X-axis line
                    showgrid=False  # Removes X-axis grid lines
                ),
                yaxis=dict(
                    title=f"{plot_type}",
                    linecolor="#BCCCDC",  # Sets color of X-axis line
                    showgrid=False  # Removes X-axis grid lines
                ),
            )
            filename = self.output_folder / \
                f"{plot_type.lower()}_{self.model_name}.html"
            plotly.offline.plot(
                fig, filename=str(filename))
            filename = self.output_folder / \
                f"{plot_type.lower()}_{self.model_name}.png"
            fig.write_image(filename)

        def _plot_multiple_models(score_files: List[Path], model_name: str, plot_type: str) -> None:
            """Plot per-entity Recall, Precision or F2 Score for multiple models given dataframes saved to csv during evaluation"""
            df = pd.read_csv(score_files[0])
            # combine score dataframes
            for file in score_files[1:]:
                df = pd.concat([df, pd.read_csv(file)])

            fig = px.bar(df, text_auto=".2", x='entity',
                         y=f'{plot_type.lower()}', color='model', barmode='group', title=f"Per-entity {plot_type} for {model_name}")
            fig.update_layout(barmode='group', xaxis={
                'categoryorder': 'total ascending'})
            fig.update_layout(yaxis_title=f"{plot_type}", xaxis_title="PII Entity")
            fig.update_traces(textfont_size=12, textangle=0,
                              textposition="outside", cliponaxis=False)
            fig.update_layout(
                plot_bgcolor="#FFF",
                xaxis=dict(
                    title="PII entity",
                    linecolor="#BCCCDC",  # Sets color of X-axis line
                    showgrid=False  # Removes X-axis grid lines
                ),
                yaxis=dict(
                    title="Recall",
                    linecolor="#BCCCDC",  # Sets color of X-axis line
                    showgrid=False  # Removes X-axis grid lines
                ),
            )
            fig.show()

        def save_errors(self):
            ModelError.most_common_fp_tokens(self.errors)

            for entity in self.model.entity_mapping.values():
                fps_df = ModelError.get_fps_dataframe(self.errors, entity=[entity])
                if fps_df is not None:
                    fps_df.to_csv(self.output_folder /
                                  f"{self.model_name}-{entity}-fps.csv")
                    fps_df = pd.read_csv(self.output_folder /
                                         f"{self.model_name}-{entity}-fps.csv")
                    self.generate_wordcloud(
                        fps_df, entity, error_type="fps", path_to_image=self.path_to_image)
                fns_df = ModelError.get_fns_dataframe(self.errors, entity=[entity])
                if fns_df is not None:
                    fns_df.to_csv(self.output_folder /
                                  f"{self.model_name}-{entity}-fns.csv")
                    fns_df = pd.read_csv(self.output_folder /
                                         f"{self.model_name}-{entity}-fns.csv")
                    self.generate_wordcloud(
                        fns_df, entity, error_type="fns", path_to_image=self.path_to_image)

        def generate_wordcloud(self, df, entity, error_type, path_to_image):
            text = ' '.join(str(v) for v in df['token'])
            if not text:
                return
            alice_mask = np.array(Image.open(path_to_image))
            # alice_mask = np.array(Image.open(Path(__file__).with_name(path_to_image)))
            wc = WordCloud(stopwords=["testing"], background_color="black", max_font_size=100, max_words=1000, mask=alice_mask,
                           contour_width=0)
            # generate word cloud
            wc.generate(text)
            # store to file
            wc.to_file(self.output_folder /
                       f"{self.model_name}-{entity}-{error_type}-wordcloud.png")

        def graph_most_common_tokens(self):

            def group(df):
                return df.groupby(['token', 'annotation']).size().to_frame(
                ).sort_values([0], ascending=False).head(30).reset_index()

            def generate_graph(type, type_title):
                df_loc = pd.read_csv(self.output_folder /
                                     f"{self.model_name}-LOC-{type}.csv")
                df_loc = group(df_loc)

                if "presidio" in self.model_name:
                    df_org = pd.read_csv(self.output_folder /
                                         f"{self.model_name}-NRP-{type}.csv")
                    df_org = group(df_org)
                else:
                    df_org = pd.read_csv(self.output_folder /
                                         f"{self.model_name}-ORG-{type}.csv")
                    df_org = group(df_org)
                df_person = pd.read_csv(self.output_folder /
                                        f"{self.model_name}-PERSON-{type}.csv")
                df_person = group(df_person)
                dfg = pd.concat([df_loc.head(3), df_org.head(3), df_person.head(3)])

                fig = px.histogram(dfg, x=0, y="token", orientation='h', color='annotation',
                                   title=f"Most common {type_title} for {self.model_name}")

                fig.update_layout(yaxis_title=f"count", xaxis_title="PII Entity")
                fig.update_traces(textfont_size=12, textangle=0,
                                  textposition="outside", cliponaxis=False)
                fig.update_layout(
                    plot_bgcolor="#FFF",
                    xaxis=dict(
                        title="Count",
                        linecolor="#BCCCDC",  # Sets color of X-axis line
                        showgrid=False  # Removes X-axis grid lines
                    ),
                    yaxis=dict(
                        title=f"Tokens",
                        linecolor="#BCCCDC",  # Sets color of X-axis line
                        showgrid=False  # Removes X-axis grid lines
                    ),
                )
                fig.update_layout(yaxis={'categoryorder': 'total ascending'})
                fig.write_image(self.output_folder /
                                f"{self.model_name}-most-common-{type}.png")

            generate_graph(type="fns", type_title="false negatives")
            generate_graph(type="fps", type_title="false positives")



from dataclasses import dataclass
import dataclasses
import json
from typing import Optional, List, Dict
from collections import Counter


@dataclass(eq=True)
class FakerSpan:
    """FakerSpan holds the start, end, value and type of every element replaced."""

    value: str
    start: int
    end: int
    type: str

    def __repr__(self):
        return json.dumps(dataclasses.asdict(self))


@dataclass()
class FakerSpansResult:
    """FakerSpansResult holds the full fake sentence, the original template
    and a list of spans for each element replaced."""

    fake: str
    spans: List[FakerSpan]
    template: Optional[str] = None
    template_id: Optional[int] = None

    def __str__(self):
        return self.fake

    def __repr__(self):
        return json.dumps(dataclasses.asdict(self))

    def toJSON(self):
        spans_dict = json.dumps([dataclasses.asdict(span) for span in self.spans])
        return json.dumps(
            {
                "fake": self.fake,
                "spans": spans_dict,
                "template": self.template,
                "template_id": self.template_id,
            }
        )

    @classmethod
    def fromJSON(cls, json_string):
        """Load a single FakerSpansResult from a JSON string."""
        json_dict = json.loads(json_string)
        converted_spans = []
        for span_dict in json.loads(json_dict['spans']):
            converted_spans.append(FakerSpan(**span_dict))
        json_dict['spans'] = converted_spans
        return cls(**json_dict)

    @classmethod
    def count_entities(cls, fake_records: List["FakerSpansResult"]) -> Counter:
        count_per_entity_new = Counter()
        for record in fake_records:
            for span in record.spans:
                count_per_entity_new[span.type] += 1
        return count_per_entity_new.most_common()

    @classmethod
    def load_privy_dataset(cls, filename: str) -> List["FakerSpansResult"]:
        """Load a dataset of FakerSpansResult from a JSON file."""
        with open(filename, "r", encoding="utf-8") as f:
            return [cls.fromJSON(line) for line in f.readlines()]

    @classmethod
    def update_entity_types(cls, dataset: List["FakerSpansResult"], entity_mapping: Dict[str, str]):
        """Replace entity types using a translator dictionary."""
        for sample in dataset:
            # update entity types on spans
            for span in sample.spans:
                span.type = entity_mapping[span.type]
            # update entity types on the template string
            for key, value in entity_mapping.items():
                sample.template = sample.template.replace(
                    "{{" + key + "}}", "{{" + value + "}}")



from .model_mock import (
    IdentityTokensMockModel,
    FiftyFiftyIdentityTokensMockModel,
    MockTokensModel,
    MockModel
)

__all__ = [
    "IdentityTokensMockModel",
    "FiftyFiftyIdentityTokensMockModel",
    "MockTokensModel",
    "MockModel"
]



from .data_objects import FakerSpan, FakerSpansResult
from .span_generator import SpanGenerator
from .record_generator import RecordGenerator
from .records_faker import RecordsFaker
from .providers import (
    NationalityProvider,
    OrganizationProvider,
    UsDriverLicenseProvider,
    IpAddressProvider,
    AddressProviderNew,
    PhoneNumberProviderNew,
    AgeProvider,
)

__all__ = [
    "SpanGenerator",
    "FakerSpan",
    "FakerSpansResult",
    "RecordGenerator",
    "NationalityProvider",
    "OrganizationProvider",
    "UsDriverLicenseProvider",
    "IpAddressProvider",
    "AddressProviderNew",
    "PhoneNumberProviderNew",
    "AgeProvider",
    "RecordsFaker",
]



from pathlib import Path
from typing import List, Optional

import requests
from spacy.training.converters import conll_ner_to_docs
from tqdm import tqdm

from presidio_evaluator import InputSample
from presidio_evaluator.dataset_formatters import DatasetFormatter


class CONLL2003Formatter(DatasetFormatter):
    def __init__(
        self,
        files_path=Path("../../data/conll2003").resolve(),
        glob_pattern: str = "*.*",
    ):
        self.files_path = files_path
        self.glob_pattern = glob_pattern

    @staticmethod
    def download(
        local_data_path=Path("../../data/conll2003").resolve(),
        conll_gh_path="https://raw.githubusercontent.com/glample/tagger/master/dataset/",
    ):

        for fold in ("eng.train", "eng.testa", "eng.testb"):
            fold_path = conll_gh_path + fold
            if not local_data_path.exists():
                local_data_path.mkdir(parents=True)

            dataset_file = Path(local_data_path, fold)
            if dataset_file.exists():
                print("Dataset already exists, skipping download")
                return

            response = requests.get(fold_path)
            dataset_raw = response.text
            with open(dataset_file, "w") as f:
                f.write(dataset_raw)
            print(f"Finished writing fold {fold} to {local_data_path}")

        print("Finished downloading CoNNL2003")

    def to_input_samples(self, fold: Optional[str] = None) -> List[InputSample]:
        files_found = False
        input_samples = []
        for i, file_path in enumerate(self.files_path.glob(self.glob_pattern)):
            if fold and fold not in file_path.name:
                continue

            files_found = True
            with open(file_path, "r", encoding="utf-8") as file:
                text = file.readlines()

            text = "".join(text)

            output_docs = conll_ner_to_docs(
                input_data=text, n_sents=None, no_print=True
            )
            for doc in tqdm(output_docs, f"Processing doc for file {file_path.name}"):
                input_samples.append(InputSample.from_spacy_doc(doc=doc))

        if not files_found:
            raise FileNotFoundError(
                f"No files found for pattern {self.glob_pattern} and fold {fold}"
            )

        return input_samples


if __name__ == "__main__":
    conll_formatter = CONLL2003Formatter()
    train_samples = conll_formatter.to_input_samples(fold="train")
    print(train_samples[:5])



from typing import List, Optional, Dict

from presidio_analyzer import EntityRecognizer
from presidio_analyzer.nlp_engine import NlpEngine

from presidio_evaluator import InputSample
from presidio_evaluator.models import BaseModel
from presidio_evaluator.span_to_tag import span_to_tag


class PresidioRecognizerWrapper(BaseModel):
    """
    Class wrapper for one specific PII recognizer
    To evaluate the entire set of recognizers, refer to PresidioAnaylzerWrapper
    :param recognizer: An object of type EntityRecognizer (in presidio-analyzer)
    :param nlp_engine: An object of type NlpEngine, e.g. SpacyNlpEngine (in presidio-analyzer)
    :param entities_to_keep: List of entity types to focus on while ignoring all the rest.
    Default=None would look at all entity types
    :param with_nlp_artifacts: Whether NLP artifacts should be obtained
        (faster if not, but some recognizers need it)
    """

    def __init__(
        self,
        recognizer: EntityRecognizer,
        nlp_engine: NlpEngine,
        entities_to_keep: List[str] = None,
        labeling_scheme: str = "BILUO",
        with_nlp_artifacts: bool = False,
        entity_mapping: Optional[Dict[str, str]] = None,
        verbose: bool = False,
    ):

        super().__init__(
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            labeling_scheme=labeling_scheme,
            entity_mapping=entity_mapping,
        )
        self.with_nlp_artifacts = with_nlp_artifacts
        self.recognizer = recognizer
        self.nlp_engine = nlp_engine

    #
    def __make_nlp_artifacts(self, text: str):
        return self.nlp_engine.process_text(text, "en")

    #
    def predict(self, sample: InputSample) -> List[str]:
        nlp_artifacts = None
        if self.with_nlp_artifacts:
            nlp_artifacts = self.__make_nlp_artifacts(sample.full_text)

        results = self.recognizer.analyze(
            sample.full_text, self.entities, nlp_artifacts
        )
        starts = []
        ends = []
        tags = []
        scores = []
        for res in results:
            if not res.start:
                res.start = 0
            starts.append(res.start)
            ends.append(res.end)
            tags.append(res.entity_type)
            scores.append(res.score)
        response_tags = span_to_tag(
            scheme="IO",
            text=sample.full_text,
            starts=starts,
            ends=ends,
            tags=tags,
            tokens=sample.tokens,
            scores=scores,
        )
        return response_tags



import collections
import json
import logging
import os
from pathlib import Path
from typing import List, Optional
import xmltodict
from tqdm import tqdm

from presidio_evaluator import InputSample
from presidio_evaluator.data_objects import Span
from presidio_evaluator.dataset_formatters import DatasetFormatter

logger = logging.getLogger("presidio-research")


class I2B22014Formatter(DatasetFormatter):
    def __init__(
        self,
        files_path="../../data/i2b2/2014/testing-PHI-Gold-fixed",
    ):
        self.files_path = files_path

    @staticmethod
    def _create_span(item):
        span = Span(
            entity_type=item["@TYPE"],
            entity_value=item["@text"],
            start_position=int(item["@start"]),
            end_position=int(item["@end"]),
        )
        return span

    def to_input_samples(self, folder: Optional[str] = None) -> List[InputSample]:
        input_samples = []
        if folder:
            self.files_path = folder
        logger.info(f"Parsing files in {self.files_path}")

        for root, dirs, files in os.walk(self.files_path):
            for file in tqdm(files, desc="Reading files..."):
                spans = []
                filename = os.path.join(root, file)
                xml_content = open(filename, "r").read()

                ordered_dict = xmltodict.parse(
                    xml_input=xml_content, strip_whitespace=False
                )
                data = dict(ordered_dict["deIdi2b2"])
                text = data["TEXT"]
                tags = data["TAGS"]
                for key, value in tags.items():
                    if key == "#text":
                        continue
                    if type(value) is collections.OrderedDict:
                        span = self._create_span(value)
                        if self._span_and_text_are_identical(span, text):
                            raise ValueError(f"Mismatch on file {file}, span {span}")

                        spans.append(span)
                    else:
                        for sub in value:
                            span = self._create_span(sub)
                            if self._span_and_text_are_identical(span, text):
                                raise ValueError(
                                    f"Mismatch on file {file}, span {span}"
                                )
                            spans.append(span)
                input_samples.append(
                    InputSample(full_text=text, spans=spans, create_tags_from_span=True)
                )

            return input_samples

    @staticmethod
    def dataset_to_json(input_path, output_path):

        formatter = I2B22014Formatter(files_path=input_path)
        train_samples = formatter.to_input_samples()
        json_dataset = [example.to_dict() for example in train_samples]

        logger.info(f"Saving dataset to json. Path: {output_path}")
        with open("{}".format(output_path), "w+", encoding="utf-8") as f:
            json.dump(json_dataset, f, ensure_ascii=False, indent=4)

    @staticmethod
    def _span_and_text_are_identical(span: Span, text: str) -> bool:
        """
        There are mismatches between annotations and text.

        To verify that the annotation matches the text,
        there is a need to ignore cases where the annotations are different
        due to whitespaces, newline or other cases.
        """

        # fmt: off
        return span.entity_value != text[
            span.start_position : span.end_position
        ].replace("Michael \n\nOgrady", "Michael   Ogrady")\
            .replace("Yehuda \n\nLanders", "Yehuda   Landers")\
            .replace("February\n\n2086", "February  2086")\
            .replace("31,\n\n", "31,  ")\
            .replace("pa\nPic", "paPic")\
            .replace("est\nBra", "estBra")\
            .replace("er \n30", "er 30")\
            .replace("pr\n29", "pr29")\
            .replace(" \n ", "  ")\
            .replace("\n\n", "")\
            .replace("\n", " ")\
            .replace("&", "and")
        # fmt: on


if __name__ == "__main__":

    # Data is assumed to be on the data folder (repo root) under i2b2/2014
    # train 1
    input_path1 = Path("../../data/i2b2/2014/training-PHI-Gold-Set1")
    output_path1 = Path("../../data/i2b2/2014/training-PHI-Gold-Set1.json")
    I2B22014Formatter.dataset_to_json(input_path1, output_path1)

    # train 2
    input_path2 = Path("../../data/i2b2/2014/training-PHI-Gold-Set2")
    output_path2 = Path("../../data/i2b2/2014/training-PHI-Gold-Set2.json")
    I2B22014Formatter.dataset_to_json(input_path2, output_path2)

    # test
    input_path3 = Path("../../data/i2b2/2014/testing-PHI-Gold-fixed")
    output_path3 = Path("../../data/i2b2/2014/testing-PHI-Gold-fixed.json")
    I2B22014Formatter.dataset_to_json(input_path3, output_path3)



from typing import Union, Dict, List

from faker import Faker
import pandas as pd

from presidio_evaluator.data_generator.faker_extensions import RecordGenerator


class RecordsFaker(Faker):
    def __init__(self, records: Union[pd.DataFrame, List[Dict]], **kwargs):
        if isinstance(records, pd.DataFrame):
            records = records.to_dict(orient="records")

        record_generator = RecordGenerator(records=records)
        super().__init__(generator=record_generator, **kwargs)



from typing import List, Optional, Dict

import spacy

from presidio_evaluator import InputSample, span_to_tag, tokenize
from presidio_evaluator.data_objects import PRESIDIO_SPACY_ENTITIES

try:
    import spacy_stanza
    import stanza
except ImportError:
    print("stanza and spacy_stanza are not installed")
from presidio_evaluator.models import SpacyModel


class StanzaModel(SpacyModel):
    """
    Class wrapping Stanza models, using spacy_stanza.

    :param model: spaCy Language object representing a stanza model
    :param model_name: Name of model, e.g. "en"
    :param entities_to_keep: List of entities to predict on
    :param verbose: Whether to print more
    :param labeling_scheme: Whether to return IO, BIO or BILUO tags
    :param entity_mapping: Mapping between input dataset entities and entities expected by the model
    """

    def __init__(
        self,
        model: spacy.language.Language = None,
        model_name: str = "en",
        entities_to_keep: List[str] = None,
        verbose: bool = False,
        labeling_scheme: str = "BIO",
        entity_mapping: Optional[Dict[str, str]] = PRESIDIO_SPACY_ENTITIES,
    ):

        if not model and not model_name:
            raise ValueError("Either model_name or model object must be supplied")
        if not model:
            model = spacy_stanza.load_pipeline(
                model_name,
                processors="tokenize,pos,lemma,ner",
            )

        super().__init__(
            model=model,
            entities_to_keep=entities_to_keep,
            verbose=verbose,
            labeling_scheme=labeling_scheme,
            entity_mapping=entity_mapping,
        )

    def predict(self, sample: InputSample) -> List[str]:
        """
        Predict the tags using a stanza model.

        :param sample: InputSample with text
        :return: list of tags
        """

        doc = self.model(sample.full_text)
        if doc.ents:
            tags, texts, start, end = zip(
                *[(s.label_, s.text, s.start_char, s.end_char) for s in doc.ents]
            )

            # Stanza tokens might not be consistent with spaCy's tokens.
            # Use spacy tokenization and not stanza
            # to maintain consistency with other models:
            if not sample.tokens:
                sample.tokens = tokenize(sample.full_text)

            # Create tags (label per token) based on stanza spans and spacy tokens
            tags = span_to_tag(
                scheme=self.labeling_scheme,
                text=sample.full_text,
                starts=start,
                ends=end,
                tags=tags,
                tokens=sample.tokens,
            )
        else:
            tags = ["O" for _ in range(len(sample.tokens))]

        if len(tags) != len(sample.tokens):
            print("mismatch between input tokens and new tokens")

        return tags



from setuptools import setup, find_packages
import os.path

# read the contents of the README file
from os import path

# this_directory = path.abspath(path.dirname(__file__))
# with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
#     long_description = f.read()
#     # print(long_description)

# with open(os.path.join(this_directory, "VERSION")) as version_file:
#     __version__ = version_file.read().strip()

setup(
    name="presidio-evaluator",
    version="0.0.71",
    packages=find_packages(exclude=["tests"]),
    url="https://www.github.com/microsoft/presidio-research",
    license="MIT",
    description="PII dataset generator, model evaluator for Presidio and PII data in general",  # noqa
    package_data={
            "presidio_evaluator":
            [
                "data_generator/raw_data/*.csv",
                "data_generator/raw_data/*.yaml",
            ],
        },
    include_package_data=True,
    install_requires=[
        "presidio_analyzer",
        "presidio_anonymizer",
        "spacy>=3.0.0",
        "requests",
        "numpy",
        "pandas",
        "tqdm>=4.32.1",
        "jupyter>=1.0.0",
        "pytest>=4.6.2",
        "haikunator",
        "schwifty",
        "faker",
        "sklearn_crfsuite",
        "python-dotenv",
    ],
)



from typing import List

import pandas as pd

from presidio_evaluator.data_generator.faker_extensions.data_objects import FakerSpansResult

try:
    from flair.data import Corpus, Sentence
    from flair.datasets import ColumnCorpus
    from flair.embeddings import (
        TokenEmbeddings,
        WordEmbeddings,
        StackedEmbeddings,
        FlairEmbeddings,
        TransformerWordEmbeddings,
    )
    from flair.models import SequenceTagger
    from flair.trainers import ModelTrainer

except ImportError:
    print("Flair is not installed")

from presidio_evaluator import InputSample
from flair.optim import LinearSchedulerWithWarmup
import torch

from os import path


class FlairTrainer:
    """
    Helper class for training Flair models
    """

    @staticmethod
    def to_flair_row(text: str, pos: str, label: str) -> str:
        """
        Turn text, part of speech and label into one row.
        :return: str
        """
        #  note: removed part of speech tag (pos) from the row because nonsensical for protocol trace data?
        # return "{} {} {}".format(text, pos, label)
        return "{} {}".format(text, label)

    def to_flair(self, df: pd.DataFrame, outfile: str = "flair_train.txt") -> None:
        """Translate a pd.DataFrame to a flair dataset."""
        sentence = 0
        flair = []
        for row in df.itertuples():
            if row.sentence != sentence:
                sentence += 1
                flair.append("")
            else:
                flair.append(self.to_flair_row(row.text, row.pos, row.label))

        if outfile:
            with open(outfile, "w", encoding="utf-8") as f:
                for item in flair:
                    f.write("{}\n".format(item))

    def create_flair_corpus(
        self, train_samples_path, test_samples_path, val_samples_path, to_bio=False
    ):
        """
        Create a flair Corpus object and saive it to train, test, validation files.
        :param train_samples_path: Path to train samples
        :param test_samples_path: Path to test samples
        :param val_samples_path: Path to validation samples
        :return:
        """
        train_samples = train_samples_path
        # train_samples = InputSample.read_dataset_json(train_samples_path)
        # train_samples = FakerSpansResult.load_privy_dataset(train_samples_path)
        train_tagged = [sample for sample in train_samples if len(sample.spans) > 0]
        print(
            f"Kept {len(train_tagged)} train samples after removal of non-tagged samples"
        )
        train_data = InputSample.create_conll_dataset(
            train_tagged, to_bio=to_bio)
        self.to_flair(train_data, outfile="flair_train.txt")

        test_samples = test_samples_path
        # test_samples = InputSample.read_dataset_json(test_samples_path)
        # test_samples = FakerSpansResult.load_privy_dataset(test_samples_path)
        test_data = InputSample.create_conll_dataset(
            test_samples, to_bio=to_bio)
        self.to_flair(test_data, outfile="flair_test.txt")

        # val_samples = FakerSpansResult.load_privy_dataset(val_samples_path)
        # val_samples = InputSample.read_dataset_json(val_samples_path)
        val_samples = val_samples_path
        val_data = InputSample.create_conll_dataset(
            val_samples, to_bio=to_bio)
        self.to_flair(val_data, outfile="flair_val.txt")

    @staticmethod
    def read_corpus(data_folder: str):
        """
        Read Flair Corpus object.
        :param data_folder: Path with files
        :return: Corpus object
        """
        # columns = {0: "text", 1: "pos", 2: "ner"}
        columns = {0: "text", 1: "ner"}
        corpus = ColumnCorpus(
            data_folder,
            columns,
            train_file="flair_train.txt",
            test_file="flair_val.txt",
            dev_file="flair_test.txt",
            in_memory=False,
        )
        return corpus

    @staticmethod
    def train_with_flair_embeddings(corpus, add_unk=False, checkpoint="", fast=False):
        """
        Train a Flair model
        :param corpus: Corpus object
        :return:
        """
        print("Corpus: ", corpus)

        # 2. what tag do we want to predict?
        tag_type = "ner"

        # 3. make the tag dictionary from the corpus
        # tag_dictionary = corpus.make_tag_dictionary(tag_type=tag_type)
        if add_unk:
            tag_dictionary = corpus.make_label_dictionary(label_type=tag_type)
        else:
            tag_dictionary = corpus.make_label_dictionary(
                label_type=tag_type, add_unk=False)
        print("Tag dictionary: ", tag_dictionary)

        # 4. initialize embeddings
        if fast:
            embedding_types: List[TokenEmbeddings] = [
                WordEmbeddings("glove"),
                FlairEmbeddings("news-forward-fast"),
                FlairEmbeddings("news-backward-fast"),
            ]
        else:
            embedding_types: List[TokenEmbeddings] = [
                WordEmbeddings("glove"),
                FlairEmbeddings("news-forward"),
                FlairEmbeddings("news-backward"),
            ]

        embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embedding_types)

        # 5. initialize sequence tagger
        tagger = SequenceTagger(
            hidden_size=256,
            embeddings=embeddings,
            tag_dictionary=tag_dictionary,
            tag_type=tag_type,
            use_crf=True,
        )

        # 6. initialize trainer
        trainer = ModelTrainer(tagger, corpus)
        if checkpoint:
            if fast:
                path = "resources/taggers/privy-flert-ner-fast/checkpoint.pt"
            else:
                path = "resources/taggers/privy-flert-ner/checkpoint.pt"
            trained_model = SequenceTagger.load(path)
            trainer.resume(
                model=trained_model,
            )
        else:
            if fast:
                path = "resources/taggers/privy-flert-ner-fast"
            else:
                path = "resources/taggers/privy-flert-ner"
            trainer.train(
                path,
                learning_rate=0.1,
                mini_batch_size=32,
                max_epochs=150,
                checkpoint=True,
            )

        sentence = Sentence("I am from Jerusalem")
        # run NER over sentence
        tagger.predict(sentence)

        print(sentence)
        print("The following NER tags are found:")

        # iterate over entities and print
        for entity in sentence.get_spans("ner"):
            print(entity)

    @staticmethod
    def train_with_transformers(corpus, add_unk=False, mini_batch_size=1, embeddings="roberta-base", checkpoint="", max_epochs=20):
        # xlm-roberta-base is the multilingual version of roberta-base
        # roberta-large is the large version of roberta-base
        """
        Train a Flair model
        :param corpus: Corpus object
        :return:
        """
        print(corpus)

        # 2. what tag do we want to predict?
        tag_type = "ner"

        # 3. make the tag dictionary from the corpus
        if add_unk:
            tag_dictionary = corpus.make_label_dictionary(label_type=tag_type)
        else:
            tag_dictionary = corpus.make_label_dictionary(
                label_type=tag_type, add_unk=False)
        print(tag_dictionary)

        # 4. initialize fine-tuneable transformer embeddings WITH document context
        # previously xlm-roberta-large
        embedding_types: List[TokenEmbeddings] = [TransformerWordEmbeddings(model=embeddings,
                                                                            layers="-1",
                                                                            subtoken_pooling="first",
                                                                            fine_tune=True,
                                                                            use_context=True,
                                                                            )]

        embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embedding_types)

        # 5. initialize bare-bones sequence tagger (no CRF, no RNN, no reprojection)
        tagger = SequenceTagger(hidden_size=256,
                                embeddings=embeddings,
                                tag_dictionary=tag_dictionary,
                                tag_type='ner',
                                use_crf=False,
                                use_rnn=False,
                                reproject_embeddings=False,
                                )

        # 6. initialize trainer
        trainer: ModelTrainer = ModelTrainer(tagger, corpus)

        if checkpoint:
            trained_model = SequenceTagger.load(
                "resources/taggers/privy-flair-transformers/checkpoint.pt")
            trainer.resume(
                model=trained_model,
                # learning_rate=5.0e-6,
                # mini_batch_size=mini_batch_size,
                # max_epochs=20,
                # optimizer=torch.optim.AdamW,
                # scheduler=LinearSchedulerWithWarmup,
                # warmup_fraction=0.1,
                # use_final_model_for_eval=True,
                # decoder_lr_factor=1.0,
                # mini_batch_chunk_size=1,  # remove this parameter to speed up computation if you have a big GPU
                # checkpoint=True,
            )

        # from torch.optim.lr_scheduler import OneCycleLR

        # 7. run fine-tuning
        else:
            trainer.fine_tune('resources/taggers/privy-flair-transformers',
                              learning_rate=5.0e-6,
                              mini_batch_size=mini_batch_size,
                              max_epochs=max_epochs,
                              #   scheduler=OneCycleLR,
                              mini_batch_chunk_size=1,  # remove this parameter to speed up computation if you have a big GPU
                              #   weight_decay=0.,
                              checkpoint=True,
                              )

        sentence = Sentence("I am from Jerusalem")
        # run NER over sentence
        tagger.predict(sentence)

        print(sentence)
        print("The following NER tags are found:")

        # iterate over entities and print
        for entity in sentence.get_spans("ner"):
            print(entity)


if __name__ == "__main__":
    train_samples = "../../data/train_Dec-19-2021.json"
    test_samples = "../../data/test_Dec-19-2021.json"
    val_samples = "../../data/validation_Dec-19-2021.json"

    trainer = FlairTrainer()
    trainer.create_flair_corpus(train_samples, test_samples, val_samples)

    corpus = trainer.read_corpus("")
    trainer.train_with_flair_embeddings(corpus)



from .dataset_formatter import DatasetFormatter
from .conll_formatter import CONLL2003Formatter
from .i2b2_formatter import I2B22014Formatter

__all__ = ["DatasetFormatter", "CONLL2003Formatter", "I2B22014Formatter"]
